/******************************************************************************
 * $Id: module.c,v 1.10 2000/04/08 22:31:57 jimb Exp $
 * Author		: Jim Brooks	<jim@jimbrooks.org>
 * Description	: Functions for managing source code modules (ie files).
 ******************************************************************************/

#include "header.h"

/******************************************************************************
 * Return a pointer to a buffer that holds a source code module (ie file).
 *
 * If the module has not been loaded, then this will load it.
 *
 * NO NEED TO FREE BUFFER MEMORY DURING LIFE OF PROGRAM.
 *
 * Returns : Pointer to memory buffer allocated by ** g_malloc() **
 *           -or- NULL upon failure.
 *
 * Parms   : pModuleName
 *           Name of module.
 *
 *         : pModuleSize (out)
 *           Size of module (undefined if module could not be loaded).
 ******************************************************************************/
gpointer ModuleBuffer( gchar* pModuleName,
                       long*  pModuleSize /* out */ )
{
   P_HASH_ITEM_MODULE	pHashItem;

  /*
   * If the sought module has been loaded into a buffer,
   * then it will have a hash item that has a non-NULL pointer to the buffer.
   */
   pHashItem = (P_HASH_ITEM_MODULE)g_hash_table_lookup( pHashModules,
                                                        (gconstpointer)pModuleName );
   if ( pHashItem  &&  pHashItem->pModuleBuffer )
   {
     /*
      * Output the pointer to the buffer, and the size of the buffer.
      */
      g_return_val_if_fail( pHashItem->pModuleBuffer, NULL );
#if 0
     /* A buffer can legitimately be empty */
      g_assert( pHashItem->moduleSize );
#endif
      *pModuleSize = pHashItem->moduleSize;
      return pHashItem->pModuleBuffer;
   }

  /*
   * The module has never been loaded, so load it now (from a file).
   * Begin by creating a hash item struct.
   */
   pHashItem = (P_HASH_ITEM_MODULE)g_malloc( sizeof(HASH_ITEM_MODULE) );
   g_return_val_if_fail( pHashItem, NULL );
#ifdef HYPERSRC_PEDANTIC
   g_return_val_if_fail( sizeof(HASH_ITEM_MODULE), NULL );
#endif

  /*
   * Load the file into a buffer.
   * Consequently, fill the hash item struct.
   */
   pHashItem->pModuleBuffer = LoadFile( pModuleName,
                                        pModuleSize /* out */ );
   if ( pHashItem->pModuleBuffer == NULL )
   {
     /*
      * Oops, could not open the module.
      */
      PrintStatusbar( "Could not open module" );
      return NULL;
   }
   pHashItem->moduleSize = *pModuleSize;

  /*
   * Insert/replace the hash item in the hash table.
   */
   g_hash_table_insert( pHashModules,
                        (gpointer)pModuleName,
                        (gpointer)pHashItem );

  /*
   * (The hash item struct will remain allocated.)
   */

  /*
   * Return a pointer to the buffer that holds the module.
   * (The size of the buffer was output to "pModuleSize".)
   */
   return pHashItem->pModuleBuffer;
}

/******************************************************************************
 * Returns a hash value from a key.
 ******************************************************************************/
guint HashModuleKey( gconstpointer pKey )
{
#if 1
   return g_str_hash( pKey );
#else
   static hashModuleKey = 0;

   return ++hashModuleKey;
#endif
}

/******************************************************************************
 * The comparsion function for the module hash.
 ******************************************************************************/
gint HashModuleCmp( gconstpointer pName1,
                    gconstpointer pName2 )
{
#if 1
   return g_str_equal( pName1,
                       pName2 );
#else
   return ! strcmp( (char*)pName1,
                    (char*)pName2 );
#endif
}
