/******************************************************************************
 * $Id: tags.c,v 1.76 2000/07/22 06:15:37 jimb Exp $
 * Author		: Jim Brooks	<jim@jimbrooks.org>
 * Description	: A set of functions for managing source code tags.
 ******************************************************************************/

#include <unistd.h>

#include "header.h"

/*
 * Local structs/vars.
 */
#ifdef HYPERSRC_THREADS
typedef struct
{
   const char*	pCtags;
   long			ctagsSize;
   GList*		pGList;
} tagsSubset_t;
static tagsSubset_t		tagsSubset = { NULL, 0, NULL };
#endif /* HYPERSRC_THREADS */

/*****************************************************************************
 * Process ctags output from a file, resulting in a GList of TAG structs.
 *
 * Parms   : pGlistTags (OUT/GLOBAL)
 *           GList of TAG structs.
 *****************************************************************************/
void ProcessCtagsFromFile( char* pFilename )
{
   char*	pCtags;
   long		ctagsSize;

  /*
   * Create a list of tags from ctags.
   */
   pCtags = ReadCtagsFromFile( &ctagsSize, /* out */
                               pTagsFilename );

#ifdef HYPERSRC_THREADS
   pGlistTags = CreateTagsGlistThreaded( pCtags, ctagsSize );
#else
   pGlistTags = CreateTagsGlist( pCtags, ctagsSize );
#endif

   return;
}

/*****************************************************************************
 * Process ctags output from stdin, resulting in a GList of TAG structs.
 *
 * Parms   : pGlistTags (OUT/GLOBAL)
 *           GList of TAG structs.
 *****************************************************************************/
void ProcessCtagsFromStdin( void )
{
   char*	pBuf;
   int      bytesRead;
   int      bytesReadMore;
   GList*	pGlistPiece = NULL;

  /*
   * Allocate enough memory for a buffer that might be > BUFSIZ.
   */
   pBuf = g_malloc( BUFSIZ * 10 );
   g_assert( pBuf );

  /*
   * This loop grabs pieces of ctags output, from stdin, and passes each piece
   * to CreateTagsGlist().  Each partial GList is appended to what will be
   * the final GList.
   */
   for (;;)
   {
     /*
      * Read the next BUFSIZ piece of ctags lines from stdin.
      * Break if end of stdin.
      */
      bytesRead = read( STDIN_FILENO, pBuf, BUFSIZ );
      if ( ! bytesRead ) break;

     /*
      * Rarely, if only one byte has been read, then read a second byte.
      * This lets the followng loop efficiently assume at least two bytes were read.
      */
      if ( bytesRead == 1 )
      {
         bytesReadMore = read( STDIN_FILENO, &pBuf[bytesRead], 1 );
         if ( ! bytesReadMore ) goto skip;
         ++bytesRead;
      }

     /*
      * Ctags lines can not be passed truncated, so unless no more input,
      * and if the last char(s) weren't EOL, try to read more chars until
      * an EOL is encountered.
      */
      while (    ! IsEolUnix(  &pBuf[bytesRead-1] )
              && ! IsEolMsdos( &pBuf[bytesRead-2] ) ) /* <-- assumes at least 2 bytes were read */
      {
         bytesReadMore = read( STDIN_FILENO, &pBuf[bytesRead], 1 );
         if ( ! bytesReadMore ) break;
         ++bytesRead;
      }

      skip:

     /*
      * Truncate buffer (though not strictly necessary).
      */
      pBuf[bytesRead] = 0;

     /*
      * Process this piece (which has untruncated lines).
      * If this is not the first piece, then append this GList into the final GList.
      */
      if ( pGlistTags == NULL )
      {
         pGlistTags = CreateTagsGlist( pBuf, bytesRead );
      }
      else
      {
         pGlistPiece = CreateTagsGlist( pBuf, bytesRead );
         g_list_concat( pGlistTags, pGlistPiece );
      }

     /*
      * Continue loop.
      */
   }

  /*
   * Were any ctag lines processed?
   */
   if ( pGlistTags == NULL )
       g_error( "No ctags lines were processed, bailing out!" );

   g_free( pBuf );
   return;
}

/*****************************************************************************
 * Spawns two threads to simultaneously create GLists from ctags input.
 *****************************************************************************/
#ifdef HYPERSRC_THREADS
GList* CreateTagsGlistThreaded( char* pCtags,
                                long  ctagsSize )
{
   const char*	pPivot;
   GList*		pGListTags = NULL;

   pthread_t		pthread;
   pthread_attr_t	pthread_attr;

  /*
   * To avoid overhread, use a single thread if only CPU or ctags input is small.
   */
   if ( SysCpus() == 1  ||  ctagsSize < 10000 )
      return CreateTagsGlist( pCtags, ctagsSize );

  /*
   * Partition the ctags input into halves that will be processed by two threads.
   * Find the start of a line that is approx. in the middle.
   */
   pPivot = NextLine( (const char*) pCtags + (ctagsSize >> 1),
                      ctagsSize >> 1 );

  /*
   * Launch a thread to process the second half.
   */
   tagsSubset.pCtags	= pPivot;
   tagsSubset.ctagsSize	= (pCtags + ctagsSize) - pPivot;

   pthread_attr_init( &pthread_attr );
   pthread_create( &pthread, &pthread_attr, ThreadCreateTagsGlist, (void*)&tagsSubset );

  /*
   * This thread will process the first half.
   */
   pGListTags = CreateTagsGlist( pCtags, pPivot - pCtags );

  /*
   * Wait until other thread finishes.
   */
Msg( "CreateTagsGlistThreaded() - waiting on spawned thread \n" );
   pthread_join( pthread, NULL );
Msg( "CreateTagsGlistThreaded() - no longer waiting \n" );

  /*
   * Merge the two GLists.
   */
   return g_list_concat( pGListTags, tagsSubset.pGList );
}
#endif /* HYPERSRC_THREADS */

/*****************************************************************************
 * This thread calls CreateTagsGlist() to process a part of the ctags input.
 * This is a thread function scheduled by pthread_create().
 *****************************************************************************/
#ifdef HYPERSRC_THREADS
void* ThreadCreateTagsGlist( void* p1 )
{
   tagsSubset_t* pTagsSubset = (tagsSubset_t*)p1;

Msg( "ThreadCreateTagsGlist() - thread start \n" );

   pTagsSubset->pGList = CreateTagsGlist( (char*)pTagsSubset->pCtags,
                                          pTagsSubset->ctagsSize );
Msg( "ThreadCreateTagsGlist() - thread end \n" );

   pthread_exit( NULL );
}
#endif /* HYPERSRC_THREADS */

/******************************************************************************
 * Create a GList (double linked list) from ctags output.
 *
 * Returns : Returns GList.
 *
 * Parms   : pCtags
 *           Pointer to raw ctags output.
 *
 *         : ctagsSize
 *           Size of ctags output.
 *
 *         : pHashModules (GLOBAL)
 ******************************************************************************/
GList* CreateTagsGlist( char* pCtags,
                        long  ctagsSize )
{
   const char*	pLine = pCtags;
   int			lines = 0;
   GList*		pGlist;

  /*
   * Ensure the tags glist pointer is NULL.
   * (The GLIB list functions will create a list if the passed list pointer is NULL.)
   */
   pGlist = NULL;

  /*
   * Decompose each line in the ctags output.
   * Add each decomposed part to a TAG struct.
   */
   for (;;)
   {
     /*
      * Stop if there are too many tags.
      */
      if ( lines++ > maxTags )
      {
         SetStatusbarInitialMsg( "There were too many tags, stopped tag processing (use -max-tags to override)." );
         break;
      }

     /*
      * Is this line empty?
      */
      if ( IsLineEmpty( pLine ) )
         break;

     /*
      * Add this ctags line to a list.
      * This list will be passed to the list widget.
      */
      if ( AddCtagsLineToTagsGlist( pLine, &pGlist ) == FUNC_OK )
      {
         P_HASH_ITEM_MODULE	pHashItem;

        /*
         * Insert/replace the hash item of the module into a hash.
         * Assign NULL to the pointer to the module buffer,
         * to indicate that it is not there yet.
         */
         pHashItem = (P_HASH_ITEM_MODULE)g_malloc( sizeof(HASH_ITEM_MODULE) );
         pHashItem->pModuleBuffer = NULL;
         pHashItem->moduleSize    = 0;
         g_hash_table_insert( pHashModules,
                              (gpointer) ((PTAG)(g_list_last( pGlist )->data))->pModule,
                              (gpointer) pHashItem );
      }
      else
      {
        /*
         * Oops, unable to process this ctags line.
         * This is considered a recoverable error (albeit unexpected)
         * only if previous ctags lines have been processed successfully,
         * which is indicated if the list does contains some elements.
         */
         if ( pGlist == NULL )
            g_error( "Could not process any ctags lines, bailing out!" );

         SetStatusbarInitialMsg( "Could not process a ctags line, proceeding anyway." );
      }

     /*
      * Move to next line in ctags.
      */
      pLine = NextLine( pLine, MAX_CTAGS_LINE_LEN );

     /*
      * Break if no more lines in ctags.
      */
      if (    pLine == NULL
           || (pLine - pCtags) >= ctagsSize ) break;
   }

#if 0
#ifdef HYPERSRC_DEBUG
   g_print( "Tag lines processed = %d \n", lines );
#endif
#endif

   return pGlist;
}

/******************************************************************************
 * Resort the tag glist according to one of: tag name, tag type, or module.
 *
 * The resort is accomplished by creating a new glist in order to use
 * g_list_insert_sort() to do the sorting work.
 *
 * Parms   : sort
 *           One of SORT_*.
 *
 *         : pGlistTags (IN/OUT/GLOBAL)
 *           The current tags glist will be freed and this will be repointed
 *           to the new sorted glist.
 *
 *         : sort1 (OUT/GLOBAL)
 *
 *         : sort2 (OUT/GLOBAL)
 *
 *         : sort3 (OUT/GLOBAL)
 ******************************************************************************/
void ResortTagsGlist( int sort )
{
   GList*	pGlistTagsNew = NULL;	/* NULL in order for glib to create the new glist */

  /*
   * Assign the sort vars appropriately.
   */
   switch ( sort )
   {
      case SORT_TAG:
         sort1    = SORT_TAG;
         sort2    = SORT_TYPE;
         sort3    = SORT_MODULE;
         break;

      case SORT_TYPE:
         sort1    = SORT_TYPE;
         sort2    = SORT_TAG;
         sort3    = SORT_MODULE;
         break;

      case SORT_MODULE:
         sort1    = SORT_MODULE;
         sort2    = SORT_TYPE;
         sort3    = SORT_TAG;
         break;

      default:
         g_assert( 0 );
   }

  /*
   * Prepare for displaying the completion percentage of sorted tags
   * in the statusbar.
   */
   tagsSorted			  = 0;
   tagsSortedPercentShown = 0;

  /*
   * Pass each elem (TAG struct) to a callback that adds each elem
   * from the current glist into a new one which is sorted as desired.
   * "pGlistTagsNew" will be passed indirectly to the callback.
   *
   * The callback will update the sort completion percentage in the statusbar.
   */
   g_list_foreach( pGlistTags,
                   ResortTagsGlistCallback,
                   (gpointer)&pGlistTagsNew );

  /*
   * Free the old glist.
   * (Free ONLY the glist ITSELF.  The TAGS structs must remain allocated.)
   */
   g_list_free( pGlistTags );

  /*
   * Update the pointer to the tags glist.
   */
   pGlistTags = pGlistTagsNew;

   return;
}

/******************************************************************************
 * Add an element from the current tags glist into a new sorted glist.
 *
 * Parms   : pTag
 *           Pointer to a TAG struct (contained by the elem of the glist).
 *           glib names this "data".
 *
 *         : ppGlistTagsNew (IN/OUT)
 *           Pointer^2 to the new tags glist.
 *           glib names this "user_data".
 ******************************************************************************/
void ResortTagsGlistCallback( gpointer pTag,
                              gpointer ppGlistTagsNew )
{
   int	tags;

  /*
   * Add this TAG struct from the current tags glist into the new one.
   * ResortTagsGlistCompare() will do up to 3 levels of comparisons
   * of the TAG struct members.
   */
   *(gpointer*)ppGlistTagsNew = g_list_insert_sorted( *(GList**)ppGlistTagsNew,
                                                      pTag,
                                                      ResortTagsGlistCompare );
  /*
   * Determine if need to update the completion percentage that
   * is shown in the statusbar.
   * Don't bother if there are few tags.
   */
   tags = g_list_length( pGlistTags );
   if ( tags > 100 )
   {
      int percent;

     /*
      * Count this sorted tag.
      * Recompute the completion percentage.
      * If the percentage has increased by 5%, then update the statusbar.
      * "tagsSortedPercentShown" holds the previous percentage being shown.
      */
      percent = (int) ( (float)(++tagsSorted) / (float)tags * 100 );
      if ( percent - tagsSortedPercentShown >= 5 )
      {
         char msg[] = "Tags sorted: __%.";
        /*
         * Record the percentage to be shown in the statusbar.
         */
         tagsSortedPercentShown = percent;

        /*
         * Show the latest completion percentage in the statusbar.
         */
         if ( percent < 100 )
         {
            msg[13] = '0' + percent / 10;
            msg[14] = '0' + percent % 10;
         }
         else
         {
            msg[13] = '9';
            msg[14] = '9';
         }

         /*
          * Clear the statusbar when the last tag is sorted,
          * else update the statusbar.
          */
         if ( tagsSorted == tags )
            MostlyClearStatusbar();
         else
            PrintStatusbar( msg );
      }
   }

   return;
}

/******************************************************************************
 * For g_list_insert_sorted(), compare the names of the modules in two TAG structs,
 * using up to 3 levels of comparisons.
 *
 * Parms   : sort1 (GLOBAL)
 *
 *         : sort2 (GLOBAL)
 *
 *         : sort3 (GLOBAL)
 ******************************************************************************/
gint ResortTagsGlistCompare( gconstpointer pTag1,
                             gconstpointer pTag2 )
{
   gint		rv;

  /*
   * Try the first-level sort.  Return if the tag fields are inequal.
   */
   rv = CompareTags( (const PTAG)pTag1,
                     (const PTAG)pTag2,
                     sort1 );
   if ( rv ) return rv;

  /*
   * Try the second-level sort.  Return if the tag fields are inequal.
   */
   rv = CompareTags( (const PTAG)pTag1,
                     (const PTAG)pTag2,
                     sort2 );
   if ( rv ) return rv;

  /*
   * Last, try the third-level sort.
   */
   rv = CompareTags( (const PTAG)pTag1,
                     (const PTAG)pTag2,
                     sort3 );
   return rv;
}

/******************************************************************************
 * Compare the members of two TAG structs.
 *
 * Returns : Result of strcmp().
 *
 * Parms   : pTag1
 *
 *         : pTag2
 *
 *         : member
 *           Which member of the TAG struct, use one of SORT_*.
 ******************************************************************************/
gint CompareTags( const PTAG pTag1,
                  const PTAG pTag2,
                  int        member )
{
   switch ( member )
   {
      case SORT_TAG:
         return (gint)strcmp( pTag1->pName,
                              pTag2->pName );

      case SORT_TYPE:
         return (gint)strcmp( pTag1->pType,
                              pTag2->pType );

      case SORT_MODULE:
         return (gint)strcmp( pTag1->pModule,
                              pTag2->pModule );

      default:
         g_assert( 0 );
   }

   return 0; /* shutup compiler */
}

/******************************************************************************
 * Add a line from ctags to a list of TAG structs.
 *
 * Parms   : pLine
 *           Pointer to start of a line in ctags.
 *
 *         : ppGList (out)
 *           Pointer to a GList pointer that is a list of TAG structs.
 *
 * Returns : FUNC_OK or FUNC_FAIL
 *
 * This is an example of a line from "ctags -x":
 *
 *  tag:    type:       line:  module:     (ignored:)
 * "main    function    14     dummy.c     int main( int argc, char* argv[] )"
 ******************************************************************************/
FUNC_X AddCtagsLineToTagsGlist( const char* pLine,
                                GList**     ppGlist )
{
   struct TAG*	pTag	= NULL;
   const char*	pWord	= pLine;
   int			sizeWord;

  /*
   * Create a TAG struct (to be added to the list after it is successfully filled).
   */
   pTag = AllocTagStruct();

  /*
   * Process the tag field.
   */
   sizeWord = ParseWord( pWord, &pWord, MAX_CTAGS_FIELD_LEN );
   if ( sizeWord > 0 )
   {
      pTag->pName = CreateString( sizeWord );
      CopyString( pTag->pName, pWord, sizeWord );
   }
   else
      return FUNC_FAIL;
   pWord = PastWord( pWord, sizeWord );

  /*
   * Process the type field.
   */
   sizeWord = ParseWord( pWord, &pWord, MAX_CTAGS_FIELD_LEN );
   if ( sizeWord > 0 )
   {
      pTag->pType = CreateString( sizeWord );
      CopyString( pTag->pType, pWord, sizeWord );

     /*
      * Is this a recognized ctags tag type?
      */
      if ( CheckTagType( pTag->pType ) != FUNC_OK )
      {
        /*
         * Tag was not immediately recognized.
         * Account for a case in which Exuberant ctags outputs a tag as two words,
         * eg: "operator !=".  If so, pTag->pName = "operator", pTag->pType "!=".
         * "operator !=" is type "function".
         */
         if ( FindSubstr( pTag->pName, "operator" ) )
         {
            char* pTmp;

           /*
            * Append the "x" to yield "operator x".
            * But first, pTag->pName must be reallocated (expanded).
            */
            pTmp = CreateString( StrLen( pTag->pName ) + 1 + StrLen( pTag->pType ) );
            CopyString( pTmp, pTag->pName, StrLen( pTag->pName ) );
            DeleteString( pTag->pName );
            pTag->pName = pTmp;

            strcat( pTag->pName, " " );
            strcat( pTag->pName, pTag->pType );

           /*
            * Probably, the type field that was parsed was actually the second word of "operator X".
            * Parse the next word which should actually be the type field.
            */ 
            pWord = PastWord( pWord, sizeWord );
            sizeWord = ParseWord( pWord, &pWord, MAX_CTAGS_FIELD_LEN );
            if ( sizeWord > 0 )
            {
              /*
               * For a second time, grab the tag type, try to identify it.
               */
               DeleteString( pTag->pType );
               pTag->pType = CreateString( sizeWord );

               CopyString( pTag->pType, pWord, sizeWord );
               if ( CheckTagType( pTag->pType ) != FUNC_OK )
                  g_error( "Could not recognize tag type '%s'", pTag->pType );
            }

           /*
            * Successfully parsed "operator X".
            */
         }
         else
         {
           /*
            * Tag is not "operator X" and is not recognized.
            */
            g_error( "Could not recognize tag type '%s'", pTag->pType );
         }
      }
   }
   else
      return FUNC_FAIL;
   pWord = PastWord( pWord, sizeWord );

  /*
   * Process the line number field.
   */
   sizeWord = ParseWord( pWord, &pWord, MAX_CTAGS_FIELD_LEN );
   if ( sizeWord > 0 )
   {
     /*
      * Copy the line number in its string form.
      */
      pTag->pLine = CreateString( sizeWord );
      CopyString( pTag->pLine, pWord, sizeWord );

     /*
      * Try to convert this field into an integer.
      */
      pTag->lineInt = atoi( pWord );
     /*
      * The result should be > 0.
      */
      if ( pTag->lineInt == 0 ) return FUNC_FAIL;
   }
   else
      return FUNC_FAIL;
   pWord = PastWord( pWord, sizeWord );

  /*
   * Process the module (ie file) field.
   */
   sizeWord = ParseWord( pWord, &pWord, MAX_CTAGS_FIELD_LEN );
   if ( sizeWord > 0 )
   {
      pTag->pModule = CreateString( sizeWord );
      CopyString( pTag->pModule, pWord, sizeWord );

     /*
      * Does this look like a module?
      */
      /* TODO */
   }
   else
      return FUNC_FAIL;
   pWord = PastWord( pWord, sizeWord );

  /*
   * Add the TAG struct to the list.
   */
   *ppGlist = g_list_append( *ppGlist, (gpointer)pTag );
   g_return_val_if_fail( *ppGlist, FUNC_FAIL );

   return FUNC_OK;
}

/******************************************************************************
 * Allocate a TAG struct.
 ******************************************************************************/
struct TAG* AllocTagStruct( void )
{
  /*
   * The TAG struct is initially all clear.
   */
   return g_malloc0( sizeof(struct TAG) );
}

/******************************************************************************
 * Free a TAG struct.
 ******************************************************************************/
void FreeTagStruct( struct TAG* pTag )
{
  /*
   * TODO
   */
}

/******************************************************************************
 * Assign the members of TAG struct.
 * Memory for each member will allocated here.
 ******************************************************************************/
void FillTagStruct( struct TAG* pTag,
                    gchar*      pName,
                    gint        lineNum,
                    gchar*      pType,
                    gchar*      pModule )
{
   /* TAG.pName */
   pTag->pName = CreateString( StrLen(pName) );
   CopyString( pTag->pName, pName, StrLen(pName) );

   /* TAG.pType */
   pTag->pType = CreateString( StrLen(pType) );
   CopyString( pTag->pType, pType, StrLen(pType) );

   /* TAG.pLine */
   pTag->pLine = CreateString( 10 );
   sprintf( pTag->pLine, "%d", lineNum );

   /* TAG.lineInt */
   pTag->lineInt = lineNum;

   /* TAG.pModuleName */
   pTag->pModule = CreateString( StrLen(pModule) );
   CopyString( pTag->pModule, pModule, StrLen(pModule) );

   return;
}

/******************************************************************************
 * Read ctags -x output from a file.
 *
 * Returns : Returns pointer to ctags output (or abends).
 *
 * Parms   : pCtagsSize (out)
 *           Size of ctags output.
 *
 *           pFilename
 ******************************************************************************/
char* ReadCtagsFromFile( long* pCtagsSize, /* out */
                         char* pFilename )
{
   gpointer	pBuffer;

   pBuffer =  LoadFile( pFilename,
                        pCtagsSize /* out */ );
   if ( pBuffer == NULL ) { g_error( "\nCould not open tags file. \n" ); }

   return pBuffer;
}

/******************************************************************************
 * Ensure that a tag type (extracted from ctags) is recognizable.
 ******************************************************************************/
FUNC_X CheckTagType( const char* pTagType )
{

#ifndef HYPERSRC_PEDANTIC
   return FUNC_OK;
#endif

  /*
   * C types:
   */
   if ( FindSubstr( pTagType, "func" ) )		return FUNC_OK;
   if ( FindSubstr( pTagType, "variable" ) )	return FUNC_OK;
   if ( FindSubstr( pTagType, "define" ) )		return FUNC_OK;
   if ( FindSubstr( pTagType, "macro" )	)		return FUNC_OK;
   if ( FindSubstr( pTagType, "struct" ))		return FUNC_OK;
   if ( FindSubstr( pTagType, "union" )	)		return FUNC_OK;
   if ( FindSubstr( pTagType, "member" ) )		return FUNC_OK;
   if ( FindSubstr( pTagType, "enum" ) )		return FUNC_OK;
   if ( FindSubstr( pTagType, "typedef" ) )		return FUNC_OK;
   if ( FindSubstr( pTagType, "prototype" )	)	return FUNC_OK;

  /*
   * C++ types:
   */
   if ( FindSubstr( pTagType, "class" )	) return FUNC_OK;

  /*
   * Assembly types:
   */
   if ( FindSubstr( pTagType, "label" ) ) return FUNC_OK;

  /*
   * M4 types:
   */
   if ( FindSubstr( pTagType, "m4_macro" ) ) return FUNC_OK;

  /*
   * TODO: other languages?
   */

  /*
   * Is this a non-tag?
   */
   if ( FindSubstr( pTagType, "(NONE)" ) ) return FUNC_OK;

  /*
   * This tag type was not recognized.
   */
   return FUNC_FAIL;
}

/*****************************************************************************
 * Goto the tag which is under the cursor (of the text widget).
 *
 * Invoked when the menu item /Goto/Tag is selected.
 *****************************************************************************/
void GotoTagUnderCursor( void )
{
   GString*		pGstr					= NULL;
   struct TAG*	pTagCurrentPosition		= NULL;
   gchar*		pText					= NULL;
   guint		textLen;
   guint		cursorIdx;
   gchar		tagName[MAX_TAG_NAME_LEN];
   int			row;

  /*
   * Ensure the text widget is created and holds text.
   */
   if ( pWidgetText == NULL  ||  ! gtk_text_get_length( GTK_TEXT(pWidgetText) ) )
   {
      PrintStatusbar( "Text widget is empty" );
      return;
   }

  /*
   * Freeze text widget, extract text.
   */
   gtk_text_freeze( GTK_TEXT(pWidgetText) );
   pText = gtk_editable_get_chars( GTK_EDITABLE(pWidgetText),
                                   0,
                                   -1 );
   textLen = gtk_text_get_length( GTK_TEXT(pWidgetText) );
   if ( pText == NULL || textLen > MAX_TEXT_BUFFER_LEN ) goto thaw_abort;

  /*
   * Get index of cursor.
   */
   cursorIdx = TextCursorIdx();
   if ( cursorIdx < 0  ||  cursorIdx > textLen ) goto thaw_abort;

  /*
   * Scoop word at cursor.
   * Return if cursor is not over a word.
   */
   pGstr = ScoopWordAtCursor( MAX_TAG_NAME_LEN,
                              pText,
                              textLen,
                              cursorIdx );
   if ( pGstr == NULL )
   {
      PrintStatusbar( "Please move the cursor over a word that might be a tag." );
      goto thaw_quit;
   }

  /*
   * Thaw text widget.
   */
   gtk_text_thaw( GTK_TEXT(pWidgetText) );

  /*
   * Create a TAG struct to define the current position,
   * and prepend it into the tags history.
   * This tag is needed so that the user can return to previous cursor positions
   * (via Prev Tag button).
   */
   pTagCurrentPosition = AllocTagStruct();
   FillTagStruct( pTagCurrentPosition,
                  "(recorded position)",
                  LineIdx2LineNum( pText, cursorIdx ),
                  "(N/A)",
                  moduleLoadedIntoTextWidget );
   PrependTagHistory( pTagCurrentPosition);

  /*
   * Select the row in the clist which contains the tag.
   * For SearchClistWidget(), create a string to hold the tag name.
   */
   CopyString( tagName, pGstr->str, pGstr->len );
   row = SearchClistWidget( tagName,
                            pWidgetClistTags,
                            widgetTagsClistRows,
                            widgetTagsClistSelectedRow,
                            TRUE, /* exact match */
                            TRUE, /* be quiet */
                            TRUE, /* OK to update notebook */
                            TRUE  /* do select row */ );

  /*
   * If the search failed then remove the tag of the current position
   * because since no hyperlink was done, the current position won't change.
   */
   if ( row == ROW_NONE)
   {
      UndoPrependTagHistory();
      PrintStatusbarThirds( "'", tagName, "' is not tagged" );

     /* TODO FreeTagStruct() */
   }

  /*
   * Free mem, return.
   */
quit:
   if ( pText ) g_free( pText );
   if ( pGstr ) g_string_free( pGstr, TRUE );
   return;

  /*
   * Oops.
   */
thaw_abort:
   g_assert( 0 );
thaw_quit:
   gtk_text_thaw( GTK_TEXT(pWidgetText) );
   goto quit;
}

/*****************************************************************************
 * Return true if a char can be part of a tag.
 *****************************************************************************/
int IsTagChar( gchar c )
{
   return    isalnum( c )
          || (c == '_') ;
}
